#include <stdio.h>
#include <sys/types.h>
#include "chrono.h"

static struct timeval sub(struct timeval tv1, struct timeval tv2);
static struct timeval add(struct timeval tv1, struct timeval tv2);
static struct timeval chr_round(struct timeval tv);

/* the chrono function handles all chronometer functionalities */
/* start, stop, lap, get current value and status */
int chrono(chr_command_t command, struct timeval *tv, struct timeval *lap) {
	struct timeval zero = { 0, 0 };
	static struct timeval previous_lap = { 0, 0 };
	static struct timeval offset = { 0, 0 };
	static struct timeval initial = { 0, 0 };
	static struct timeval now;
	static int running = 0;

	switch(command) {
	case CHR_RESET :
		running = 0;
		offset = zero;
		previous_lap = zero;
		break;
	case CHR_LAP :
		if (!running) {
			*tv = offset;
			*lap = zero;
		} else {
			gettimeofday(&now, NULL);
			/* get absolute time */
			*tv = add(offset, sub(now, initial));
			*lap = sub(chr_round(*tv), previous_lap);
			previous_lap = chr_round(*tv);
		}
		break;
	case CHR_START_STOP :
		if (!running) { /* do start */
			gettimeofday(&initial, NULL);
			running = 1;
		} else { /* do stop */
			gettimeofday(&now, NULL);
			*tv = add(offset, sub(now, initial));
			*lap = sub(chr_round(*tv), previous_lap);
			previous_lap = chr_round(*tv);
			offset = add(offset, sub(now, initial));
			running = 0;
		}
		break;
	case CHR_POLL : /* get current values */
		if (!running) {
			*tv = offset;
			*lap = zero; 
		} else {
			gettimeofday(&now, NULL);
			*tv = add(offset, sub(now, initial));
			*lap = sub(chr_round(*tv), previous_lap);
		}
		break;
	case CHR_RUNNING : /* get status */
		break;
	default :
		fprintf(stderr, "Error in chrono: invalid value (%d) for command\n", command);
		*tv = zero;
		*lap = zero;
		break;
	}
	return running;
}

/* get an int value for the 1/100 seconds */
/* this is used so that lap value are consistent with
 * the values of the chronometer */
static struct timeval chr_round(struct timeval tv) {
	struct timeval result;
	result = tv;
	result.tv_usec /= 1e4;
	result.tv_usec *= 1e4;
	return result;
}

static struct timeval add(struct timeval tv1, struct timeval tv2) {
	struct timeval result;
	result.tv_sec = tv1.tv_sec + tv2.tv_sec;
	result.tv_usec = tv1.tv_usec + tv2.tv_usec;
	if (result.tv_usec>=1e6) {
		result.tv_sec += 1;
		result.tv_usec -= 1e6;
	}
	return result;
}

static struct timeval sub(struct timeval tv1, struct timeval tv2) {
	struct timeval result;
	if ((tv1.tv_sec>tv2.tv_sec) || ((tv1.tv_sec==tv2.tv_sec) && (tv1.tv_usec>=tv2.tv_usec))) { /* tv1 >= tv2 */
		result.tv_sec = tv1.tv_sec - tv2.tv_sec;
		result.tv_usec = tv1.tv_usec - tv2.tv_usec;
		if (result.tv_usec<0) {
			result.tv_usec += 1e6;
			result.tv_sec -= 1;
		}
	} else { /* tv1 < tv2 */
		result = sub(tv2, tv1);
		result.tv_sec = - result.tv_sec;
		result.tv_usec = - result.tv_usec;
	}
	return result;
}
